import argparse
import logging
import pathlib
from collections.abc import Iterator
from typing import Optional

import grpc
import numpy as np
from google.protobuf.duration_pb2 import Duration
from google.protobuf.json_format import MessageToJson
from phonexia.grpc.common.core_pb2 import Audio, RawAudioConfig, TimeRange
from phonexia.grpc.technologies.audio_quality_estimation.v1.audio_quality_estimation_pb2 import (
    EstimateRequest,
    EstimateResponse,
)
from phonexia.grpc.technologies.audio_quality_estimation.v1.audio_quality_estimation_pb2_grpc import (
    AudioQualityEstimationStub,
)
from soundfile import SoundFile


def time_to_duration(time: float) -> Optional[Duration]:
    if time is None:
        return None
    duration = Duration()
    duration.seconds = int(time)
    duration.nanos = int((time - duration.seconds) * 1e9)
    return duration


def make_request(
    file: str,
    start: Optional[float],
    end: Optional[float],
    use_raw_audio: bool,
) -> Iterator[EstimateRequest]:
    time_range = TimeRange(start=time_to_duration(start), end=time_to_duration(end))
    chunk_size = 1024 * 100
    if use_raw_audio:
        with SoundFile(file) as r:
            raw_audio_config = RawAudioConfig(
                channels=r.channels,
                sample_rate_hertz=r.samplerate,
                encoding=RawAudioConfig.AudioEncoding.PCM16,
            )
            for data in r.blocks(blocksize=r.samplerate, dtype="float32"):
                int16_info = np.iinfo(np.int16)
                data_scaled = np.clip(
                    data * (int16_info.max + 1), int16_info.min, int16_info.max
                ).astype("int16")
                yield EstimateRequest(
                    audio=Audio(
                        content=data_scaled.flatten().tobytes(),
                        raw_audio_config=raw_audio_config,
                        time_range=time_range,
                    )
                )
                time_range = None
                raw_audio_config = None

    else:
        with open(file, mode="rb") as fd:
            while chunk := fd.read(chunk_size):
                yield EstimateRequest(audio=Audio(content=chunk, time_range=time_range))
                time_range = None


def write_result(
    audio_path: str, response: EstimateResponse, output_file: Optional[str] = None
) -> None:
    json_response = MessageToJson(
        message=response,
        indent=2,
        ensure_ascii=False,
        preserving_proto_field_name=True,
        always_print_fields_with_no_presence=True,
    )
    if output_file:
        logging.info(f"Writing results of the estimation to '{output_file}'")
        with open(output_file, "w") as f:
            f.write(json_response)
    else:
        print(json_response)


def estimate(
    channel: grpc.Channel,
    file: str,
    output_file: Optional[str],
    start: Optional[float],
    end: Optional[float],
    metadata: Optional[list],
    use_raw_audio: bool,
) -> None:
    logging.info(f"Estimate quality from '{file}'")
    stub = AudioQualityEstimationStub(channel)
    response = stub.Estimate(
        make_request(file, start, end, use_raw_audio),
        metadata=metadata,
    )
    write_result(file, response, output_file)


def existing_file(file: str) -> str:
    if not pathlib.Path(file).exists():
        raise argparse.ArgumentError(argument=None, message=f"File {file} does not exist")
    return file


def main():
    parser = argparse.ArgumentParser(
        description="Audio quality estimation gRPC client. analyzing the quality of audio recordings."
    )
    parser.add_argument(
        "-H",
        "--host",
        type=str,
        default="localhost:8080",
        help="Server address, default: localhost:8080",
    )
    parser.add_argument(
        "-l",
        "--log_level",
        type=str,
        default="error",
        choices=["critical", "error", "warning", "info", "debug"],
    )
    parser.add_argument(
        "--metadata",
        metavar="key=value",
        nargs="+",
        type=lambda x: tuple(x.split("=")),
        help="Custom client metadata",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=pathlib.Path,
        default=None,
        required=False,
        help="Output file with the results. If not set, print to standard output.",
    )

    parser.add_argument("--use_ssl", action="store_true", help="Use SSL connection")
    parser.add_argument("--start", type=float, help="Audio start time")
    parser.add_argument("--end", type=float, help="Audio end time")
    parser.add_argument("--use_raw_audio", action="store_true", help="Send a raw audio")
    parser.add_argument("file", type=existing_file, help="input audio file")

    args = parser.parse_args()

    if args.start is not None and args.start < 0:
        raise ValueError("Parameter 'start' must be a non-negative float.")

    if args.end is not None and args.end <= 0:
        raise ValueError("Parameter 'end' must be a positive float.")

    if args.start is not None and args.end is not None and args.start >= args.end:
        raise ValueError("Parameter 'end' must be larger than 'start'.")

    logging.basicConfig(
        level=args.log_level.upper(),
        format="[%(asctime)s.%(msecs)03d] [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    try:
        logging.info(f"Connecting to {args.host}")
        if args.use_ssl:
            with grpc.secure_channel(
                target=args.host, credentials=grpc.ssl_channel_credentials()
            ) as channel:
                estimate(
                    channel=channel,
                    file=args.file,
                    output_file=args.output,
                    start=args.start,
                    end=args.end,
                    metadata=args.metadata,
                    use_raw_audio=args.use_raw_audio,
                )
        else:
            with grpc.insecure_channel(target=args.host) as channel:
                estimate(
                    channel=channel,
                    file=args.file,
                    output_file=args.output,
                    start=args.start,
                    end=args.end,
                    metadata=args.metadata,
                    use_raw_audio=args.use_raw_audio,
                )

    except grpc.RpcError as e:
        logging.error(f"RPC failed: {e}")  # noqa: TRY400
        exit(1)
    except Exception:
        logging.exception("Unknown error")
        exit(1)


if __name__ == "__main__":
    main()
